from __future__ import annotations

import inspect
from itertools import islice
import textwrap
from typing import Callable, get_args, Literal, NamedTuple, Union

from .domain import domain_expr, Domain, DomainCoercible, Env


# Property = ForAll Property
#          | Exists Property
#          | Predicate
#
# Posibles críticas a esta definición de `Property`:
#
#   - `Property = Predicate` es un caso muy extremo de propiedad.
#
#   - Nada impide que el programado construya mal una propiedad. En concreto
#     puede terner variables libres (en nuestro caso, sin cuantificar).
#     Sin embargo el conjunto de errores que puede cometer un programador es
#     infinito. No podemos evitarlos todos.
#
# Las variables libres en un predicado pueden no ser tal cosa si incorporamos
# conceptos como las `fixtures` de `pytest`. En este caso son variables que
# no están ligadas en ningún cuantificador, pero están ligadas en otro contexto
# más global (y compartido por varias propiedades).


# :( property es una palabra clave en python
class QCProperty: pass


def is_qcproperty(x):
    return isinstance(x, QCProperty)


VarName = str




#---------------------------------------------------------------------------
# Expresiones de dominio con variables libres
#---------------------------------------------------------------------------
# Expresiones con variables libres. Una vez ligadas las variables, si
# evaluamos la expresión, obtenemos un dominio.
#
# Se representa como una función cuyos parámetros son las variables libres.

DomainLambda = Callable[[...], DomainCoercible]

class DomainExprWithFreeVars(NamedTuple):
    fun: DomainLambda
    unbound_vars: list[str]

    def __str__(self) -> str:
        return f"{self.unbound_vars} => {self.fun}"


def domain_expr_with_free_vars(fun: DomainLambda) -> DomainExprWithFreeVars:
    signature = inspect.signature(fun)
    unbound_vars = list(signature.parameters.keys())
    if len(unbound_vars) == 0:
        raise TypeError(f"no free variables")
    return DomainExprWithFreeVars(fun, unbound_vars)
    

def bind_and_eval(expr: Union[DomainExprWithFreeVars, Domain],
                  env: Env) -> Domain:
    if not isinstance(expr, DomainExprWithFreeVars):
        return expr
    # A la hora de construir kwargs ignoramos la variables que no
    # están en env para que el error salte al llamar a la función
    kwargs = { k: env[k] for k in expr.unbound_vars if k in env}
    return domain_expr(expr.fun(**kwargs))
    

#---------------------------------------------------------------------------
# Tipos.
# Syntax sugar y otras facilities
#---------------------------------------------------------------------------
# Parámetro de la variable en el decorador de cuantificación.
QArg = Union['DomainLambda', DomainCoercible]


def _preprocess_domain(arg: QArg) -> Domain:
    if getattr(arg, '__name__', None) == '<lambda>':
        return domain_expr_with_free_vars(arg)
    else:
        return domain_expr(arg)


class Left:
    def __init__(self, e):
        # TODO: e puede ser Env o [Env,Exception]
        #       regularizar esto
        self.e = e

    def __bool__(self):
        return False

    def __str__(self):
        return f"Left({self.e})"
    
    
# -- Either bool
CheckResult = Union[Left, Literal[True]]


#---------------------------------------------------------------------------
# Predicados
#---------------------------------------------------------------------------
# Implementación (función python) de un predicado
PredicateFun = Callable[..., bool]


# Adaptador de predicado a QC.
class Predicate(QCProperty):
    def __init__(self, pred: PredicateFun):
        self.pred = pred
        
    def __call__(self, /, env: Env) -> Iterator[CheckResult]:
        try:
            result = self.pred(**env) or Left(env)
        except Exception as e:
            result = Left((env, e))
        yield result

    def __str__(self):
        return str(self.pred.__name__)


#---------------------------------------------------------------------------
# Cuantificadores
#---------------------------------------------------------------------------
# Podríamos implementar los cuantificadores como funciones (cierres).
# Los implementamos así para incluir otras operaciones interesantes
# como p.e. `__str__`

# Todas las propiedades se "ejecutan" en un `Environment` que contiene
# las variables ligadas por cada uno de los cuantificadores.


class ForAll(QCProperty):
    def __init__(self,
                 quantifed_var: VarName,
                 domain_obj: Domain,
                 qcproperty: QCProperty,
                 n_samples: int):
        self.quantifed_var = quantifed_var
        self.domain_obj = domain_obj
        self.qcproperty = qcproperty
        self.n_samples = n_samples

    def __call__(self, /, env: Env) -> Iterator[CheckResult]:
        quantifed_var = self.quantifed_var
        if quantifed_var in env:
            raise TypeError(f"Variable {quantifed_var} is shadowed in {self}")

        domain_obj = bind_and_eval(self.domain_obj, env)
        qcproperty = self.qcproperty

        # Si el dominio está marcado como finito recorremos el dominio entero
        if domain_obj.is_exhaustible:
            domain_samples = domain_obj.exhaustible
        else:
            domain_samples = islice(domain_obj, self.n_samples)
            
        for sample in domain_samples:
            yield from qcproperty(env= {**env, quantifed_var: sample})

    def __str__(self):
        return (f"ForAll {self.quantifed_var}: {self.domain_obj}\n"
                f"{textwrap.indent(str(self.qcproperty), '  ')}")

        
class Exists(QCProperty):
    def __init__(self,
                 quantifed_var: VarName,
                 domain_obj: Domain,
                 qcproperty: QCProperty):
        if not isinstance(qcproperty, Predicate):
            # TODO: Podríamos comprobar que todos los cuantificadores dentro
            #       de este Exists son exhaustivos ?
            #       Seguramente es más sencillo añadir un parámetro opcional
            #       a la función `qc` que indique si tiene que ser exhaustive
            raise TypeError(f"This tool cannot check Exists on another quantifier")
        
        self.quantifed_var = quantifed_var
        self.domain_obj = domain_obj
        self.qcproperty = qcproperty

        
    def __call__(self, /, env: Env) -> Iterator[CheckResult]:
        quantifed_var = self.quantifed_var
        if quantifed_var in env:
            raise TypeError(f"Variable {quantifed_var} is shadowed in {self}")

        domain_obj = bind_and_eval(self.domain_obj, env)
        qcproperty = self.qcproperty

        if not domain_obj.is_exhaustible:
            raise TypeError(f"It's not possible to check existence "
                            f"in non exhaustive domain: {domain_obj}")
        
        for sample in domain_obj.exhaustible:
            if any(qcproperty(env= {**env, quantifed_var: sample})):
                yield True
                return
                
        yield Left(f"{env} -> {str(self)}")
        return

    def __str__(self):
        return f"Exists {self.quantifed_var}: {self.domain_obj} / {self.qcproperty}"
        
    
#---------------------------------------------------------------------------
# decoradores
#---------------------------------------------------------------------------
def forall(n_samples: int= 100, **binds):
    """Decorates a predicate funcion or another decorator with a forall quantifier.

    Parameters
    ----------
    n_samples : int
        The number of samples to check.
    **binds
        The quantified variables. Actually the number of quantified
        variables is limited to 1.

    Returns
    -------
    ForAll
        An object implementing the property as a callable.
    """
    if len(binds) != 1:
        # TODO: permitir esto como "azúcar sintáctico" ?
        #       Es decir que
        #           @forall(x= ..., y= ...)
        #       sea lo mismo que
        #           @forall(x= ...)
        #           @forall(y= ...)
        #       La pregunta es si el código es más inteligible.
        raise TypeError(f"Must bind just one variable, but {len(binds)} binded")
    
    def factory(arg):
        if is_qcproperty(arg):
            qcproperty = arg
        else:
            qcproperty = Predicate(arg)
        quantifed_var, obj = list(binds.items())[0]
        domain_obj = _preprocess_domain(obj)
        return ForAll(quantifed_var, domain_obj, qcproperty, n_samples= n_samples)

    return factory


def exists(**binds):
    """Decorates a predicate funcion or another decorator with an existencial quantifier.

    Parameters
    ----------
    **binds
        The quantified variables. Actually the number of quantified
        variables is limited to 1.

    Returns
    -------
    Exists
        An object implementing the property as a callable.

    !!! warning
        Every variable must be quantified over an exhaustible domain.

    """
    if len(binds) != 1:
        # TODO: permitir esto como "azúcar sintáctico" ?
        raise TypeError(f"Must bind just one variable, but {len(binds)} binded")

    def factory(arg):
        if is_qcproperty(arg):
            qcproperty = arg
        else:
            qcproperty = Predicate(arg)
        quantifed_var, obj = list(binds.items())[0]
        domain_obj = _preprocess_domain(obj)
        return Exists(quantifed_var, domain_obj, qcproperty)

    return factory




from __future__ import annotations

from dataclasses import dataclass, field, InitVar, replace
import inspect
from itertools import islice
import random
import string
from typing import Any, Callable, Generator, Generic, Iterable, Iterator, Optional, Protocol, Union, Sized, TypeVar
from typing import runtime_checkable

# TODO: ¿ queremos/podemos hacer que la semilla no sea una variable global del módulo ?

seed = random.randint(0, 10000)
_random = random.Random()
_random.seed(seed)


def get_seed():
    return seed


def set_seed(new_seed):
    global seed
    seed = new_seed
    _random.seed(new_seed)

    
#---------------------------------------------------------------------------
# Tipos
#---------------------------------------------------------------------------

# Actually, cualquier objeto se puede convertir en un dominio.
# Ver DomainFromIterable o DomainSingletion
DomainCoercible = Any


# Entorno. Guarda las ligaduras de variables.
Env = dict[str, DomainCoercible]


def is_domain(arg: Any) -> bool:
    return isinstance(arg, Domain)


def domain_expr(arg: DomainCoercible,
                is_exhaustible: Optional[bool]= None) -> Domain:
    """Desugar domain expressions

    Converts any expression declaring a domain into a Domain object.
    The options are:

    - If arg is already a Domain object, returns it. Parameters are
      not allowed.

    - If arg is any python iterable, returns a Domain object whose
      elements are the items in the iterable. The programmer may mark
      the domain as exhaustible when neccessary.

    - If arg is a generator function, equivalent to iterable, taking
      the elements from the generator returned by the function.

    - If arg is any other python object, returns a Domain containing
      only one element: arg.

    """
    if is_domain(arg):
        if is_exhaustible is not None:
            raise TypeError("cannot change the attribute is_exhaustible of a domain")
        return arg
    elif isinstance(arg, Iterable):
        return DomainFromIterable(arg, is_exhaustible= is_exhaustible or False)
    elif inspect.isgeneratorfunction(arg):
        return DomainFromGeneratorFun(arg,  is_exhaustible= is_exhaustible or False)
    else:
        return DomainSingleton(arg)
    # TODO: Otras formas de declarar un dominio, p.e.
    #    - tipos básicos: int, str, ...
    #    - función generadora

            
RecStepFun = Callable[['RecDomainStep'], 'Domain']


def recursive(fun: RecStepFun, **props) -> RecDomain:
    def rec_dom_factory():
        return RecDomain(fun, **props)
    return rec_dom_factory


# ---------------------------------------------------------------------------
# Implementación de dominios
# ---------------------------------------------------------------------------
"""
Conocemos tres estrategias principales para generar la lista de
samples:

- Naïf. A partir de una semilla, generar una secuencia de números
  pseudoaleatorios y usar estos números para seleccionar lo elementos
  del conjunto.

- TODO: A mayores del caso anterior, usar heurísticas para incluir
  elementos que se sabe que son "problemáticos" en las pruebas, por
  ejemplo los zeros.

- TODO: Sized generation. Los elementos se eligen en orden creciente
  de tamaño y complejidad.  
"""

T = TypeVar('T')
class Domain(Generic[T]):
    """Domain abstract class

    This is the base class for every `Domain`. It includes some
    operations like _union_.

    It also includes the method `that` to decorate a domain with extra
    funcionality. For example: limiting the number of samples that it
    will generate.

    By default, domains are non exhaustibles. The attribute
    `is_exhaustible = False`.  When a domain is or may be exhaustible,
    the corresponding class has to explicitly change the attribute
    `is_exhaustible` and, if neccessary, add a parameter to the
    `__init__` function. And, of course, implement the exhaustible
    iterator `exhaustible_iter`.

    """
    is_exhaustible: bool = False
    
    @property
    def exhaustible(self) -> Iterator[T]:
        """Returns an iterator that can be exhausted.

        """
        if not self.is_exhaustible:
            raise TypeError(f"cannot exhaust domain {self}")
        else:
            return self.exhaustible_iter()
            
    def __iter__(self) -> Iterator[T]:
        """Returns the cannonical iterator for a domain.

        This is the magic method of any python iterable. The returned
        iterator is the pbt cannonical one. This iterator will pick
        and yield random samples from the domain.

        """
        raise NotImplementedError()

    def exhaustible_iter(self) -> Iterator[T]:
        """Returns the non-cannonical, exhaustible iterator for a domain.

        The returned iterator is not the pbt cannonical one. Instead
        of picking random samples, this iterator will pick and yield
        all the elements of the domain.

        NOTE that most domains can not be exhausted because the number
        of elements is too big, generating them is too costly, or,
        most of the cases, the number of elements is infinite.

        WARNING do not use this method directly, instead use the
        property `exhaustible`.

        """
        raise NotImplementedError()
    
    def __or__(self, other: DomainCoercible) -> Domain:
        return DomainUnion(self, other)
    
    def __ror__(self, other: DomainCoercible) -> Domain:
        return DomainUnion(other, self)

    def that(self, samples_limit: int) -> Domain:
        """Decorate this domain

        Returns a decorator of this domain. The decorator will change
        the behaviour of the domain based on the given parameters.

        Parameters
        ----------
        samples_limit : int
            max number of samples the domain generates

        Returns
        -------
        Domain
            the decorated domain
        """
        return DomainLimit(self, samples_limit)


# Decorator
class DomainLimit(Domain):
    def __init__(self, delegate: Domain, samples_limit: int):
        self.delegate = delegate
        self.samples_limit = samples_limit

    def __iter__(self) -> Iterator[T]:
        return islice(self.delegate, self.samples_limit)


    

class RecDomain(Domain):
    def __init__(self, fun: RecStepFun, sub_i: int= 0, max_depth:int= 6):
        self.fun = fun
        self.sub_i = sub_i
        self.max_depth = max_depth

    def __iter__(self) -> Iterator:
        return self._samples(self.sub_i + 1)

    def _samples(self, sub_i: int) -> Iterator:
        if sub_i > self.max_depth:
            raise RecursionError()
        subdom = self.fun(lambda max_depth= 6: RecDomain(self.fun, sub_i= sub_i, max_depth= max_depth))
        yield from subdom

    def __str__(self) -> str:
        return f"RecDomain_{self.sub_i}({id(self)})"


class DomainSingleton(Domain):
    is_exhaustible = True
    
    def __init__(self, element: Any):
        self.element = element

    def __iter__(self) -> Iterator:
        element = self.element
        while True:
            yield element

    def exhaustible_iter(self) -> Iterator:
        yield self.element


class DomainFromIterable(Domain):
    def __init__(self, iterable: Iterable, is_exhaustible: bool):
        self.iterable = iterable
        self.is_exhaustible = is_exhaustible

    def __iter__(self) -> Iterator:
        if not self.is_exhaustible:
            yield from self.iterable
        else:
            # Importante: barajar los items por si acaso el número de muestras que
            # se cogen del dominio es menor que el tamaño del iterable, para evitar
            # devolver siempre los $n$ primeros.
            samples = self.iterable
            if not isinstance(samples, Sequence):
                # Por ejemplo: un generator object es un Iterable
                # Pero la función random.sample sólo admite Sequence como
                # parámetro (ver doc.)
                samples = tuple(samples)
            samples = _random.sample(samples, len(samples))
            while True:
                yield from samples

    def exhaustible_iter(self) -> Iterator:
        return iter(self.iterable)


class DomainFromGeneratorFun(Domain):
    def __init__(self, fun: Callable, is_exhaustible: bool):
        self.fun = fun
        self.is_exhaustible = is_exhaustible

    def __iter__(self) -> Iterator:
        generator = self.fun()
        if not self.is_exhaustible:
            yield from generator
        else:
            # Importante: barajar los items por si acaso el número de muestras que
            # se cogen del dominio es menor que el tamaño del iterable, para evitar
            # devolver siempre los $n$ primeros.
            samples = tuple(generator)
            samples = _random.sample(samples, len(samples))
            while True:
                yield from samples

    def exhaustible_iter(self) -> Iterator:
        return self.fun()

    

class Sublists(Domain):
    def __init__(self, l: list, is_exhaustible: bool= False):
        self.l = l
        self.is_exhaustible = is_exhaustible
        
    def __iter__(self) -> Iterator[list]:
        yield []
        n = len(self.l)
        while True:
            a = _random.randint(0, n)
            b = _random.randint(0, n)
            a, b = (a, b + 1) if a <= b else (b, a + 1)
            yield self.l[a:b]
        
    def exhaustible_iter(self) -> Iterator[list]:
        if not self.is_exhaustible:
            raise RuntimeError("Domain is not marked as exhaustible")
        yield []
        n = len(self.l)
        for a in range(n):
            for b in range(a, n):
                yield self.l[a:b+1]


@dataclass(frozen= True)
class Int(Domain):
    min_value: int = 0
    max_value: int = 10_000
    
    def __iter__(self) -> Iterator[int]:
        min_value = self.min_value
        max_value = self.max_value
        if min_value <= 0 <= max_value:
            yield 0
        while True:
            yield _random.randint(min_value, max_value)


class PyName(Domain):
    def __init__(self, min_len: Optional[int]= 1, max_len: Optional[int]= 8):
        if min_len and min_len < 1:
            raise ValueError("min len ({min_len}) must be greater than zero. No python names allowed with less than one char")
        if max_len < min_len:
            raise ValueError(f"max len ({max_len}) cannot be smaller than min len ({min_len})")
        self.min_len = min_len
        self.max_len = max_len
        
    # TODO: Generar nombres de variables más adecuados
    def __iter__(self) -> Iterator:
        n = self.max_len - self.min_len
        head_chars = ['_', *string.ascii_letters]
        tail_chars = [*head_chars, *string.digits]
        
        while True:
            chars = [
                _random.choice(head_chars),
                *_random.choices(tail_chars, k= n)
            ]
            
            yield "".join(chars)

    def __str__(self) -> str:
        return f"PyName()"
    

@dataclass(frozen= True)
class List(Domain):
    domain_obj: InitVar[DomainCoercible] = None
    min_len: Optional[int] = field(default= 0, kw_only= True)
    max_len: Optional[int] = field(default= 20, kw_only= True)
    domain: Domain = field(init= False)

    def __post_init__(self, domain_obj: DomainCoercible):
        super().__setattr__('domain', domain_expr(domain_obj))
        
    def __iter__(self) -> Iterator[list]:
        min_len = self.min_len
        max_len = self.max_len
        if min_len == 0:
            yield []
            min_len = 1
        while True:
            n = _random.randint(min_len, max_len)
            yield list(islice(self.domain, n))

    
class Tuple(Domain):
    def __init__(self, *domain_objs: tuple(DomainCoercible,...)):
        self.domains = [domain_expr(d) for d in domain_objs]

    def __iter__(self) -> Iterator:
        iterators = [iter(domain) for domain in self.domains]
        while True:
            yield tuple(next(it) for it in iterators)

    def exhaustible_iter(self) -> Iterator:
        if any(not domain.is_exhaustible for domain in self.domains):
            raise TypeError(f"not every element is exhaustible in {self}")
        iterators = [domain.exhaustible for domain in self.domains]
        # TODO: si un dominio es más grande que otro, ¿ nos vale este comportamiento
        #  o queremos implementar otros ?
        return zip(*iterators)

    def __str__(self):
        items = ", ".join(map(str, self.domains))
        return f"Tuple({items})"


class Dict(Domain):
    def __init__(self, key_domain: DomainCoercible, value_domain: DomainCoercible, min_len: int= 0, max_len: int= 20):
        self.key_domain = domain_expr(key_domain)
        self.value_domain = domain_expr(value_domain)
        self.min_len = min_len
        self.max_len = max_len

    def __iter__(self) -> Iterator:
        min_len = self.min_len
        max_len = self.max_len
        if min_len == 0:
            yield {}
            min_len = 1
        key_it = iter(self.key_domain)
        value_it = iter(self.value_domain)
        while True:
            n = _random.randint(min_len, max_len)
            yield { next(key_it): next(value_it) for _ in range(n) }

    def __str__(self):
        return f"Dict({self.key_domain}, {self.value_domain})"
        
        
class Boolean(Domain):
    is_exhaustible : bool = True

    def as_exhaustible(self, exhaustible: bool) -> Domain:
        if not exhaustible:
            raise TypeError(f"do no mark boolean as not exhaustible: {self}")
        return self
    
    def __iter__(self) -> Iterator:
        while True:
            yield _random.choice([True, False])

    def exhaustible_iter(self) -> Iterator:
        yield False
        yield True

    def __str__(self):
        return f"Boolean()"


class None_(Domain):
    is_exhaustible : bool = True

    def as_exhaustible(self, exhaustible: bool) -> Domain:
        if not exhaustible:
            raise TypeError(f"do no mark None as not exhaustible: {self}")
        return self
        
    def __iter__(self) -> Iterator:
        while True:
            yield None

    def exhaustible_iter(self) -> Iterator:
        yield None
        
    def __str__(self) -> str:
        return f"None()"
    

class DomainUnion(Domain):
    a: InitVar[Domain]
    b: InitVar[Domain]
    domains: list[Domain]= field(init= False)
    is_exhaustible: bool= field(init= False)
    
    def __init__(self, a: DomainCoercible, b: DomainCoercible):
        if isinstance(a, DomainUnion):
            a = a.domains
        else:
            a = [domain_expr(a)]
        if isinstance(b, DomainUnion):
            b = b.domains
        else:
            b = [domain_expr(b)]
        self.domains = [*a, *b]
        self.is_exhaustible = all(d.is_exhaustible for d in self.domains)

    def as_exhaustible(self, exhaustible: bool) -> Domain:
        raise TypeError(f"can not mark union as exhaustible: {self}")
            
    def __iter__(self) -> Iterator:
        iterators = [iter(domain) for domain in self.domains]
        n = len(iterators)
        while True:
            idxs = _random.sample(range(n), k= n)
            for i in idxs:
                try:
                    yield next(iterators[i])
                    break
                except RecursionError:
                    # Si alguna alternativa llega al nivel de
                    # recursión máximo, la descartamos y reiniciamos
                    # el iterador para la siguiente
                    iterators[i] = iter(self.domains[i])
            else:
                raise RecursionError(f"max recursion at {self}")

    # TODO: ¿ Cómo detectar si es suficientemente finito ?
    #       ¿ La unión de dos dominios suficientemente finitos puede dar lugar a uno no suficientemente finito ?
    
    def __str__(self):
        return " | ".join(map(str, self.domains))


class DomainPyObject(Domain):
    def __init__(self,
                 factory: Callable,
                 *domain_args: tuple(DomainCoercible,...),
                 **domain_kwargs: dict[str, DomainCoercible]):
        self.factory = factory
        self.domain_args = [domain_expr(d) for d in domain_args]
        self.domain_kwargs = { key: domain_expr(d)
                               for key, d in domain_kwargs.items() }

    def __iter__(self) -> Iterator:
        factory = self.factory
        args_iterators = [iter(domain) for domain in self.domain_args]
        kwargs_iterators = { key: iter(domain)
                             for key, domain in self.domain_kwargs.items() }
        while True:
            yield factory(*(next(it) for it in args_iterators),
                          **{ key: next(it)
                              for key, it in kwargs_iterators.items() })

    def exhaustible_iter(self) -> Iterator:
        if any(not domain.is_exhaustible for domain in self.domain_args):
            raise TypeError(f"not every element is exhaustible in {self}")
        factory = self.factory
        args_iterators = [domain.exhaustible for domain in self.domain_args]
        kwargs_iterators = { key: domain.exhaustible
                             for key, domain in self.domain_kwargs.items() }

        # TODO: si un dominio es más grande que otro, ¿ nos vale este comportamiento
        #  o queremos implementar otros ?
        while True:
            try:
                yield factory(*(next(it) for it in args_iterators),
                              **{ key: next(it)
                                  for key, it in kwargs_iterators.items() })
            except StopIteration:
                break



#!/usr/bin/env python3

from __future__ import annotations

import importlib
from pathlib import Path
import sys

from pypbt.quantifier import is_qcproperty

def run_props(file: Path) -> None:
    parent = str(file.parent)
    module_name = file.stem
    if parent not in sys.path:
        sys.path.append(parent)
    module = importlib.import_module(module_name)
    for name in dir(module):
        obj = getattr(module, name)
        if is_qcproperty(obj):
            prop = obj
            print(prop)
            for i, result in enumerate(prop(env= {}), start= 1):
                if result:
                    print(".", end= "", flush= True)
                else:
                    print("x")
                    print(f"After {i} tests" if i>1 else "After 1 test")
                    print(result)
                    break
            else:
                print()
                print(f"Passed {i} tests" if i>1 else "Passed 1 test")
            print()


def main():
    if len(sys.argv) != 2:
        cmd = Path(sys.argv[0]).name
        print(f"USAGE: {cmd} <props_file>")
        sys.exit(1)

    file = Path(sys.argv[1])
    if not file.exists():
        print(f"ERROR: File not found: {file}")
        sys.exit(1)

    run_props(file)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pypbt']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['pypbt = pypbt.runner:main']}

setup_kwargs = {
    'name': 'pypbt',
    'version': '0.1.2',
    'description': 'Library of functions to write Property Based Tests in python',
    'long_description': '# pypbt\n\nThis library implements functions to write _property based tests_ in\npython.\n\nIt started as a proof of concept. By now there are still some missing\nfeatures:\n\n  - Shrinking\n  \n  - Generation strategies\n  \n  - State machines\n  \nWe can not make any promises about them.\n\n\n## Installation\n\n```\npip install pypbt\n```\n\n## Documentation\n\nSee [documentation on readthedocs](https://pypbt.readthedocs.io/en/latest/).\n\n## Support\n\nPlease [open an issue](https://github.com/cabrero/pypbt/issues) for support.\n\n## License\n\nThe project is licensed under the LGPL license.\n',
    'author': 'David Cabrero Souto',
    'author_email': 'david.cabrero@madsgroup.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cabrero/pypbt',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
